from typing import Dict
from .AppImageProvider import AppImageProvider

# A list containing all the "Providers"
# providers: Dict[str, Provider] = { 
#     'appimage': AppImageProvider()
# }

appimage_provider = AppImageProvider()